#include "Python.h"
#include "string.h"
#include <ggi/ggi.h>


typedef struct {
  PyObject_HEAD
  ggi_visual_t vis;
  const ggi_pixelformat *pixelformat;
} VisObject;
extern PyTypeObject vis_type;
#define PyVis_Check(op) ((op)->ob_type == &vis_type)

typedef struct {
  PyObject_HEAD
  ggi_mode mode;
} ModeObject;
extern PyTypeObject mode_type;
#define PyMode_Check(op) ((op)->ob_type == &mode_type)

typedef struct {
  PyObject_HEAD
  ggi_event ev;
  PyObject* dict;
} EventObject;
extern PyTypeObject event_type;
#define PyEvent_Check(op) ((op)->ob_type == &event_type)

typedef struct {
  PyObject_HEAD
  ggi_visual_t visual;
  ggi_pixel *pixels;
  ggi_color *colors;
  int length;
} ColorObject;
extern PyTypeObject color_type;
#define PyColor_Check(op) ((op)->ob_type == &color_type)

/* Utility Functions -----------------------------------------------*/
void floatcolor(ggi_color*c, double r,  double g, double b) 
{
  c->r=(int)(r*65535);
  c->g=(int)(g*65535);
  c->b=(int)(b*65535);
}

void colorfloat(ggi_color*c, double*r, double*g,double*b) 
{
  *r=((double)c->r)/65535;
  *g=((double)c->g)/65535;
  *b=((double)c->b)/65535;
}

int pyfloat(PyObject*o,double*f) {
  if(PyInt_Check(o)) {
    *f=(double)PyInt_AS_LONG(o); 
  } else if(PyFloat_Check(o)) {
    *f=PyFloat_AS_DOUBLE(o); 
  } else {
    PyErr_SetString(PyExc_TypeError, "must supply a number!"); 
    return 0;
  }
  return 1; // as in, not NULL
}


/* End of Utility Functions ----------------------------------------*/

/* ModeObject ------------------------------------------------------*/
void mode_dealloc(ModeObject*self) {
  PyMem_DEL(self);
}

int mode_cmp(ModeObject*a,ModeObject*b) {
  if((a->mode.frames==b->mode.frames) &&
     (a->mode.virt.x==b->mode.virt.x) &&
     (a->mode.virt.y==b->mode.virt.y) &&
     (a->mode.visible.x==b->mode.visible.x) &&
     (a->mode.visible.y==b->mode.visible.y) &&
     (a->mode.size.x==b->mode.size.x) &&
     (a->mode.size.y==b->mode.size.y) &&
     (a->mode.dpp.x==b->mode.dpp.x) &&
     (a->mode.dpp.y==b->mode.dpp.y) &&
     (a->mode.graphtype==b->mode.graphtype)
     ) return 0;
  return -1; // nope, sorting will NOT work on modes.
}

PyObject* mode_getattr(ModeObject*self, char*name) {
#define coord2tuple(c) Py_BuildValue("ii",c.x, c.y)

  switch(name[0]) {
  case 'f': 
    if(strcmp(name,"frames")==0) return PyInt_FromLong(self->mode.frames);
    break;
  case 'v':
    if(strcmp(name,"visible")==0) {
      return coord2tuple(self->mode.visible);
    } else if(strcmp(name,"virt")==0) {
      return coord2tuple(self->mode.virt);
    }
    break;
  case 's':
    if(strcmp(name,"size")==0) return coord2tuple(self->mode.visible);
    break;
  case 'g':
    if(strcmp(name,"graphtype")==0) return PyInt_FromLong(self->mode.graphtype);
    break;
  case 'd':
    if(strcmp(name,"dpp")==0) {
      return coord2tuple(self->mode.dpp);
    }
    break;
  } 
  PyErr_SetString(PyExc_AttributeError, name);
  return NULL;
#undef coord2tuple
}

int mode_setattr(ModeObject*self, char*name, PyObject*val) {
  int x,y;
#define coordcheck(v) (PyTuple_Check(val) && PyArg_ParseTuple(val,"ii",&x,&y))
  switch(name[0]) {
  case 'f': 
    if(strcmp(name,"frames")==0 && PyInt_Check(val)) {
      self->mode.frames=PyInt_AsLong(val);
      return 0;
    }
    break;
  case 'v':
    if(coordcheck(val)) {
      if(strcmp(name,"visible")==0) {
	self->mode.visible.x=x;
	self->mode.visible.y=y;
	return 0;
      } else if(strcmp(name,"virt")==0) {
	self->mode.virt.x=x;
	self->mode.virt.y=y;
	return 0;
      }
    }
    break;
  case 's':
    if(strcmp(name,"size")==0 && coordcheck(val)) {
      self->mode.size.x=x;
      self->mode.size.y=y;
      return 0;
    }
    break;
  case 'g':
    if(strcmp(name,"graphtype")==0 && PyInt_Check(val)) {
      self->mode.graphtype=PyInt_AsLong(val);
      return 0;
    }
    break;
  case 'd':
    if(strcmp(name,"dpp")==0 && coordcheck(val)) {
      self->mode.dpp.x=x;
      self->mode.dpp.y=y;
      return 0;
    }
    break;
  } 
  PyErr_SetString(PyExc_AttributeError, name);
  return -1;
#undef coordcheck
}

static PyTypeObject mode_type = {
  PyObject_HEAD_INIT(NULL)
  0,
  "GGIMode",
  sizeof(ModeObject),
  0,
  (destructor)mode_dealloc,
  0,
  (getattrfunc)mode_getattr, 
  (setattrfunc)mode_setattr,
  (cmpfunc)mode_cmp,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0L,0L,0L,0L,
  "no docs."
};

PyObject* newModeFromGGIMode(ggi_mode m) {
  ModeObject*newmode;
  if((newmode=PyObject_NEW(ModeObject, &mode_type))==NULL) return NULL;
  newmode->mode=m;
  return (PyObject*)newmode;
}

PyObject* ModeConstruct(VisObject*self,PyObject*args,PyObject*keys) {
  ggi_mode m;
  ModeObject*mode;
  
  PyObject *foo;

  if(PyTuple_Size(args)!=0) {
    PyErr_SetString(PyExc_TypeError, "Mode requires only keyword args");
    return NULL;
  }

  m.frames=GGI_AUTO;
  m.graphtype=GGI_AUTO;
  m.visible.x=GGI_AUTO;
  m.visible.y=GGI_AUTO;
  m.virt.x=GGI_AUTO;
  m.virt.y=GGI_AUTO;
  m.dpp.x=GGI_AUTO;
  m.dpp.y=GGI_AUTO;
  m.size.x=GGI_AUTO;
  m.size.y=GGI_AUTO;

  if(self!=NULL) {
    int res=ggiCheckMode(self->vis, &m);
  }

  if(keys && PyDict_Check(keys)) {
#define getcoord(arg)  \
    if((foo=PyDict_GetItemString(keys, #arg))!=NULL) { \
      Py_INCREF(foo);\
      if((!PyTuple_Check(foo)) || PyTuple_Size(foo)!=2) {\
	PyErr_SetString(PyExc_TypeError, "that should be a coord, bonehead.");\
	return NULL;\
      }\
      m. ## arg ## .x = PyInt_AsLong(PyTuple_GetItem(foo,0)); \
      m. ## arg ## .y = PyInt_AsLong(PyTuple_GetItem(foo,1)); \
      Py_DECREF(foo);\
    }
#define getint(arg) \
    if((foo=PyDict_GetItemString(keys, #arg))!=NULL) { \
      Py_INCREF(foo);\
      if(!PyInt_Check(foo)) {\
	PyErr_SetString(PyExc_TypeError, "that should be an int, bucko.");\
	return NULL;\
      }\
      m. ## arg = PyInt_AsLong(PyTuple_GetItem(foo,0)); \
      Py_DECREF(foo);\
    }
    getcoord(visible);
    getcoord(virt);
    getcoord(dpp);
    getcoord(size);
    getint(graphtype);
    getint(frames);
#undef getcoord
#undef getint
  }

  mode=PyObject_NEW(ModeObject, &mode_type);
  if(mode!=NULL) mode->mode=m;

  return (PyObject*)mode;

  // TODO: stick Mode() in main module methods, and check here to see where we are being called from.
  //if it's the module, do the above. else, make a mode with those mods, do a checkmode on it, and return
  //that mode.
}


/* End of ModeObject -----------------------------------------------*/

/* ColorObject -----------------------------------------------------*/
// I'm not really impressed with ColorObjects... They could be far nicer,
// but I'm really not an OOD guy..

//these must be defined here because they're used above the type 
//declaration but their implementations must be below it.
ColorObject* newColorFromVisualAndSize(ggi_visual_t v, int n);
ColorObject* newColorFromColorAndVis(ColorObject*color,ggi_visual_t v);

void color_dealloc(ColorObject*self) {
  if(self->pixels) free(self->pixels);
  if(self->colors) free(self->colors);
  PyMem_DEL(self);
}


int singlecolorcmp(ColorObject*oa,ColorObject*ob,int i) {
  ldiv_t da,db;

  ggi_color a=oa->colors[i];
  ggi_color b=ob->colors[i];
  
  long int aa=a.r+a.g+a.b;
  long int bb=b.r+b.g+b.b;

  if(a.r==b.r && a.g==b.g && a.b==b.b) {
    return 0;
  };

  

  da=ldiv(aa,3);
  db=ldiv(bb,3);

  return da.quot-db.quot;
}

int singlepixelcmp(ColorObject*oa,ColorObject*ob,int i) {
  ldiv_t da,db;

  ggi_pixel a=oa->pixels[i];
  ggi_pixel b=ob->pixels[i];

  // should a check that a&b both have equal stdformats? no, they should always be
  // of the same visual here.

  if(a==b) {
    return 0;
  }

  //notice that i don't do anything special here.. just add up their values 
  //interpreted as a long int... this means they'd sort funny. but who wants to sort
  //color objects???
  da=ldiv(a,3);
  db=ldiv(b,3);

  return da.quot-db.quot;
}

int color_cmp(ColorObject*v,ColorObject*w) {
  int i;
  int (*singlecmp)(ColorObject*,ColorObject*,int);
  
  if (v->visual!=NULL && v->visual==w->visual) {
    singlecmp=singlepixelcmp;
  }
  else if (v->visual==w->visual) { //ie, both null
    singlecmp=singlecolorcmp;
  }
  else { //different visuals
    ColorObject*n=newColorFromColorAndVis(w,v->visual);
    w=n;    
    singlecmp=singlepixelcmp;
  }
  for (i = 0; i < v->length && i < w->length; i++) {
    int cmp = singlecmp(v,w,i);
    if (cmp != 0)
      return cmp;
  }
  return v->length - w->length;
}

PyObject* color_repr(ColorObject* self) {
  char *buf1, buf2[80];
  if(self->colors) {
    buf1=" as ggi_colors";
  } else if(self->pixels) {
    buf1=" as ggi_pixels";
  } else {
    buf1=", all fucked up.";
  }
  sprintf(buf2, "<GGIColor of length %i%s>", self->length, buf1);
  return PyString_FromString(buf2);
}

int color_seq_len(ColorObject*self) {
  return self->length;
}

PyObject *color_seq_concat(ColorObject*self,PyObject*other) {
  int size,i;
  ColorObject*newcolor;
#define other_ ((ColorObject*)other)

  if(!PyColor_Check(other)) {
    PyErr_BadArgument();
    return NULL;
  }

  if (self->visual!=other_->visual) {
    PyErr_BadArgument();
    return NULL;  // can't concat between visuals
  }

  size=self->length+other_->length;
  newcolor=newColorFromVisualAndSize(self->visual,size);
  if(newcolor==NULL) return NULL;
  if(self->visual) { // copy the ggi_pixels
    for(i=0;i<self->length;i++) newcolor->pixels[i]=self->pixels[i];
    for(i=0;i<other_->length;i++) newcolor->pixels[self->length+i]=other_->pixels[i];
  } else {  // copy the ggi_colors
    for(i=0;i<self->length;i++) newcolor->colors[i]=self->colors[i];
    for(i=0;i<other_->length;i++) newcolor->colors[self->length+i]=other_->colors[i];
  }
#undef other_
  return (PyObject*)newcolor;
}

PyObject *color_seq_repeat(ColorObject*self, int n) {
  int size,i,j;
  ColorObject*newcolor;

  if(n<0) n=0;
  size=self->length*n;
  newcolor=newColorFromVisualAndSize(self->visual,size);
  if(newcolor==NULL)return NULL;
  if(self->visual) {
    for(i=0;i<n;i++) {
      for(j=0;i<self->length;j++) {
	newcolor->pixels[i]=self->pixels[j];
      }
    }
  } else {
    for(i=0;i<n;i++) {
      for(j=0;i<self->length;j++) {
	newcolor->colors[i]=self->colors[j];
      }
    }
  }

  return (PyObject*)newcolor;
}

PyObject* color_seq_item(ColorObject*self, int i) {
  double r,g,b;
  ColorObject *co;

  if (i < 0 || i >= self->length) {
    PyErr_SetString(PyExc_IndexError,"color index out of range"); 
    return NULL;
  }  

  co=newColorFromVisualAndSize(self->visual, 1);
  if(co!=NULL) {
    if(self->visual) {
      uint8 *temp_p=(void*)self->pixels;
      temp_p+=(ggiGetPixelFormat(self->visual)->size/8)*i;
      co->pixels[0]=*(ggi_pixel*)temp_p;
    } else {
      co->colors[0]=self->colors[i];
    }
  }
  

  return (PyObject*)co;
}

PyObject* color_seq_slice(ColorObject*self, int ilow, int ihigh) {
  int i;
  ColorObject*newcolor;

  printf("color slice:asked to produce a slice from %i to %i\n", ilow, ihigh);
  printf("color slice:converting arguments...\n");

  if(ilow<0) ilow=0;
  else if (ilow>self->length) ilow=self->length;
  printf("color slice:.. %i %i\n", ilow, ihigh);
  
  if(ihigh<ilow) ihigh=ilow;
  else if(ihigh>self->length) ihigh=self->length;
  printf("color slice:.. %i %i\n", ilow, ihigh);

  newcolor=newColorFromVisualAndSize(self->visual, ihigh-ilow);
  if(newcolor==NULL) return NULL;
 
  if(self->visual) {    
    int pixel_size=ggiGetPixelFormat(self->visual)->size/8;
    uint8 *newindex=(void*)newcolor->pixels;
    uint8 *oldindex=(void*)self->pixels;
    oldindex+=ilow*pixel_size;

    printf("color slice:i have a visual, so copying pixels..\n");
    for(i=ilow;i<ihigh; i++) {
      *(ggi_pixel*)newindex=*(ggi_pixel*)oldindex;
      newindex+=pixel_size;
      oldindex+=pixel_size;
    }
  }else{
    printf("color slice:i have no visual so copying colors..\n");
    for(i=ilow;i<ihigh; i++) {
      newcolor->colors[i-ilow]=self->colors[i];
    }
  }

  printf("color slice: done.. now returning.\n\n");
  return (PyObject*)newcolor;
}

int color_seq_ass_item(ColorObject*self, int i, PyObject*value) {
  double r,g,b;
  ggi_color c;

  if(i<0 || i>self->length) {
    PyErr_SetString(PyExc_IndexError, "color assignment index out of range");
    return -1;
  }
  if(value==NULL) {
    PyErr_SetString(PyExc_IndexError, "can't del slices of colors");
    return -1;
  }


  if(PyTuple_Check(value) && PyTuple_Size(value)==3) {
#define check_float_range(f) if((f)<0 || (f)>1) { \
    PyErr_SetString(PyExc_TypeError, "need three numbers 0..1"); \
    return -1; }

    check_float_range(r);
    check_float_range(g);
    check_float_range(b);
#undef check_float_range
    floatcolor(&c, r,g,b);
    if(self->visual) {
      ggi_pixel p=ggiMapColor(self->visual, &c);
      self->pixels[i]=p;
    } else {
      self->colors[i]=c;
    }
  } else if(PyColor_Check(value)) {
    if(((ColorObject*)value)->length!=1) {
      PyErr_SetString(PyExc_TypeError, "can't expand/contract colors.");
      return -1;
    }
    if(self->visual!=NULL && self->visual==((ColorObject*)value)->visual) {
      int pixel_size=ggiGetPixelFormat(self->visual)->size/8; //really oughtta be cached..
      self->pixels[i*pixel_size]=((ColorObject*)value)->pixels[0]; //?
    } else if (self->visual==NULL) {
      self->colors[i]=((ColorObject*)value)->colors[0];
    } else {
      ColorObject*temp;
      temp=newColorFromColorAndVis((ColorObject*)value, self->visual);
      self->pixels[i]=temp->pixels[0];
      Py_DECREF(temp);
    }
  } else {
    PyErr_SetString(PyExc_TypeError, "bad arguments.");
    return -1;
  }

  

  return 0;
}

int color_seq_ass_slice(ColorObject*self,int ilow, int ihigh, PyObject*value) {
#define value_ ((ColorObject*)value)  
  int i;

  if(ilow<0) ilow=0;
  else if(ilow>self->length) ilow=self->length;
  if(ihigh<ilow) ihigh=ilow;
  else if(ihigh>self->length) ihigh=self->length;

  if(value==NULL) {
    PyErr_SetString(PyExc_IndexError, "can't del slices of colors");
    return -1;
  }
  if(!PyColor_Check(value)) {
    PyErr_SetString(PyExc_TypeError, "need another color object of same length as slice");
    return -1;
  }
  if(value_->length!=(ihigh-ilow)) {
    PyErr_SetString(PyExc_TypeError, "need another color object of same length as slice");
    return -1;
  }
  if(self->visual!=value_->visual) {//a "value" visual is a visual that's only a bit dented...
    PyErr_SetString(PyExc_TypeError, "can't assign between visuals!");
    return -1;
  }


  if(self->visual) {
    for(i=ilow;i<ihigh;i++) {
      self->pixels[i]=value_->pixels[i-ilow];
    }
  }else{
    for(i=ilow;i<ihigh;i++) {
      self->colors[i]=value_->colors[i-ilow];
    }
  }
  
  return 0;
#undef value_
}


static PySequenceMethods color_as_sequence = {
  (inquiry)color_seq_len,
  (binaryfunc)color_seq_concat,
  (intargfunc)color_seq_repeat,
  (intargfunc)color_seq_item,
  (intintargfunc)color_seq_slice,
  (intobjargproc)color_seq_ass_item,
  (intintobjargproc)color_seq_ass_slice
};

static PyTypeObject color_type = {
  PyObject_HEAD_INIT(NULL)
  0,
  "GGIColor",
  sizeof(ColorObject),
  0,
  (destructor)color_dealloc,
  (printfunc)0,
  (getattrfunc)0, 
  (setattrfunc)0,
  (cmpfunc)color_cmp,
  (reprfunc)color_repr,
  0,
  &color_as_sequence,
  0,
  (hashfunc)0,
  (ternaryfunc)0,
  (reprfunc)0,
  0L,0L,0L,0L,
  "no docs."
};

ColorObject* newColorFromColorAndVis(ColorObject*color,ggi_visual_t v) {
  int n=color->length;
  int i;
  int pixel_size=ggiGetPixelFormat(v)->size/8;
  uint8 *p1,*p2;

  ColorObject*newcolor=PyObject_NEW(ColorObject,&color_type);
  if(newcolor==NULL) return NULL; //TODO: Should i set an error here?
  newcolor->visual=v;
  newcolor->length=n;
  newcolor->pixels=NULL; newcolor->colors=NULL;

  if(v==color->visual) {
    newcolor->pixels=(ggi_pixel*)malloc(pixel_size*n);
    p1=(void*)newcolor->pixels;
    p2=(void*)color->pixels;
    for(i=0;i<n;i++) {
      *(ggi_pixel*)p1=*(ggi_pixel*)p2;
      p1+=pixel_size;
      p2+=pixel_size;
    }
    //  } else if() {
  } else { 
    newcolor->colors=(ggi_color*)malloc(sizeof(ggi_color)*n);
    for(i=0;i<n;i++) newcolor->colors[i]=color->colors[i];
  }

  return newcolor;
}

ColorObject* newColorFromVisualAndSize(ggi_visual_t v, int n) {
  ColorObject*newcolor=PyObject_NEW(ColorObject,&color_type);
  int pixel_size=ggiGetPixelFormat(v)->size/8;

  printf("newColorFromVisualAndSize asked to make color_type of size %i\n ", n);
  if(newcolor==NULL) return NULL;
  newcolor->visual=v;
  newcolor->length=n;
  newcolor->pixels=NULL; newcolor->colors=NULL;
  if(v) {
    newcolor->pixels=(ggi_pixel*)malloc(pixel_size*n);
  } else{
    newcolor->colors=(ggi_color*)malloc(sizeof(ggi_color)*n);
  }

  if(newcolor->pixels==NULL && newcolor->colors==NULL) {
    PyErr_SetString(PyExc_SystemError, "Ouf of memory allocating color\n");
    return NULL;
  }     

  return newcolor;
}

ColorObject* newColorFromNumbersAndVisual(PyObject*ro,PyObject*go,PyObject*bo,ggi_visual_t v) {
  double r,g,b;
  ggi_color c;
  ColorObject*newcolor;


  if((pyfloat(ro,&r) | pyfloat(go,&g) | pyfloat(bo,&b))==0) {
    PyErr_SetString(PyExc_TypeError, "need three numbers 0..1");
    return NULL;
  }

#define check_float_range(f) if((f)<0 || (f)>1) { \
    PyErr_SetString(PyExc_TypeError, "need three numbers 0..1"); \
    return NULL; }

  check_float_range(r);
  check_float_range(g);
  check_float_range(b);
#undef check_float_range

  floatcolor(&c, r,g,b);

  newcolor=PyObject_NEW(ColorObject,&color_type);
  if(newcolor==NULL) return NULL;
  newcolor->visual=v;
  newcolor->length=1;
  newcolor->pixels=NULL; newcolor->colors=NULL;  

#define checkmem(arg) if(!newcolor-> arg) {\
    PyErr_SetString(PyExc_SystemError, "not enough memory..somehow");\
    return NULL;}

  if(v) {
    ggi_pixel p;
    int pixel_size=ggiGetPixelFormat(v)->size/8;
    p=ggiMapColor(v, &c);  
    newcolor->pixels=(ggi_pixel*)malloc(pixel_size); // ick!
    checkmem(pixels)
    newcolor->pixels[0]=p;
  } else{
    newcolor->colors=(ggi_color*)malloc(sizeof(ggi_color)); // ick!
    checkmem(colors)
    newcolor->colors[0]=c;
  }

  return newcolor;

}

PyObject* ColorConstruct(VisObject*self, PyObject*args) {
  /*verify arguments: we take:
    a. three numbers 0..1
    b. a sequence of sequences containing three number 0..1 (not yet)
    c. nothing at all, in which case we return a color with r=g=b=0
  */

  ggi_visual_t v;
  PyObject *ro,*go,*bo;

  if(self==NULL) v=NULL;
  else v=self->vis;

  if(PyArg_ParseTuple(args, "OOO", &ro,&go, &bo)) 
    return (PyObject*)newColorFromNumbersAndVisual(ro,go,bo,v);
  //if(PyArg_ParseTuple(args, "O", &list)) 
  //  return (PyObject*)newColorFromListAndVisual(list,v);
  if(PyTuple_Size(/*(PyTupleObject*)*/args)==0) {
    ro=PyFloat_FromDouble(0);
    return (PyObject*)newColorFromNumbersAndVisual(ro,ro,ro,v);
    Py_DECREF(ro);
  }
  
  PyErr_SetString(PyExc_TypeError,
		  "Color() needs (a) three numbers 0..1 (b) a sequence of sequences of "
		  "three numbers 0..1 or (c) no arguments.");
  return NULL;
}
/* End of ColorObject ----------------------------------------------*/

/* EventObject -----------------------------------------------------*/
static PyObject* event_getattr(EventObject*self, char*name) {
  PyObject* o;

  

  if (strcmp(name,"type")==0) {
    return PyInt_FromLong(self->ev.any.type);
  }
  else if (strcmp(name,"error")==0) {
    return PyInt_FromLong(self->ev.any.error);
  }
  else if (strcmp(name,"origin")==0) {
    return PyInt_FromLong(self->ev.any.origin);
  }
  else if (strcmp(name,"target")==0) {
    return PyInt_FromLong(self->ev.any.target);
  }
  else if (strcmp(name,"time")==0) {
    return PyInt_FromLong((self->ev.any.time.tv_sec)*1000+self->ev.any.time.tv_usec);
  }
  
  switch (self->ev.any.type) {
  case evKeyPress:
  case evKeyRelease:
  case evKeyRepeat:
    if (strcmp(name,"modifiers")==0) {
      return PyInt_FromLong(self->ev.key.modifiers);
    } 
    else if (strcmp(name,"sym")==0) {
      return PyInt_FromLong(self->ev.key.sym);
    }
    else if (strcmp(name,"label")==0) {
      return PyInt_FromLong(self->ev.key.label);
    }
    else if (strcmp(name,"button")==0) {
      return PyInt_FromLong(self->ev.key.button);
    }
    break;

  case evPtrRelative:
  case evPtrAbsolute:
    if (strcmp(name,"x")==0) {
      return PyInt_FromLong(self->ev.pmove.x);
    } 
    else if (strcmp(name,"y")==0) {
      return PyInt_FromLong(self->ev.pmove.y);
    }
    else if (strcmp(name,"z")==0) {
      return PyInt_FromLong(self->ev.pmove.z);
    }
    else if (strcmp(name,"wheel")==0) {
      return PyInt_FromLong(self->ev.pmove.wheel);
    }
    break;

  case evPtrButtonPress:
  case evPtrButtonRelease:
    if (strcmp(name,"button")==0) {
      return PyInt_FromLong(self->ev.pbutton.button);
    } 
    break;

  case evExpose:
    if (strcmp(name,"x")==0) {
      return PyInt_FromLong(self->ev.expose.x);
    } 
    else if (strcmp(name,"y")==0) {
      return PyInt_FromLong(self->ev.expose.y);
    }
    else if (strcmp(name,"h")==0) {
      return PyInt_FromLong(self->ev.expose.h);
    }
    else if (strcmp(name,"w")==0) {
      return PyInt_FromLong(self->ev.expose.w);
    }    
    break;

  case evCommand:
  case evInformation:
    if (strcmp(name, "code")==0) {
      return PyInt_FromLong(self->ev.cmd.code);
    }
    /* else if (strcmp(name, "data")==0) {
       // do something snazzy..
       }
    */
    break;

  case evValRelative:
  case evValAbsolute:
    // need more info on valuators.
    break;

  default:      
  }

  return NULL;
  
}


static int event_setattr(EventObject*self, char *name, PyObject*value) {
  PyObject* o;
#define _ci {if(!PyInt_Check(value)) return 0;}
#define _i PyInt_AsLong(value)

  if (strcmp(name,"type")==0) {
    _ci;
    self->ev.any.type=_i;
    return 1;
  }
  else if (strcmp(name,"error")==0) {
    _ci;
    self->ev.any.error=_i;
    return 1;
  }
  else if (strcmp(name,"origin")==0) {
    _ci;
    self->ev.any.origin=_i;
    return 1;
  }
  else if (strcmp(name,"target")==0) {
    _ci;
    self->ev.any.target=_i;
    return 1;
  }

  /*  else if (strcmp(name,"time")==0) {
    self->ev.any.time.=_i;
    return PyInt_FromLong((self->ev.any.time.tv_sec)*1000+self->ev.any.time.tv_usec);
    }*/
  
  switch (self->ev.any.type) {
  case evKeyPress:
  case evKeyRelease:
  case evKeyRepeat:
    _ci;
    if (strcmp(name,"modifiers")==0) {
      self->ev.key.modifiers=_i;
      return 1;
    } 
    else if (strcmp(name,"sym")==0) {
      _ci;
      self->ev.key.sym=_i;
      return 1;
    }
    else if (strcmp(name,"label")==0) {
      _ci;
      self->ev.key.label=_i;
      return 1;
    }
    else if (strcmp(name,"button")==0) {
      _ci;
      self->ev.key.button=_i;
      return 1;
    }
    break;

    /* finish these later.
  case evPtrRelative:
  case evPtrAbsolute:
    if (strcmp(name,"x")==0) {
      return PyInt_FromLong(self->ev.pmove.x);
    } 
    else if (strcmp(name,"y")==0) {
      return PyInt_FromLong(self->ev.pmove.y);
    }
    else if (strcmp(name,"z")==0) {
      return PyInt_FromLong(self->ev.pmove.z);
    }
    else if (strcmp(name,"wheel")==0) {
      return PyInt_FromLong(self->ev.pmove.wheel);
    }
    break;

  case evPtrButtonPress:
  case evPtrButtonRelease:
    if (strcmp(name,"button")==0) {
      return PyInt_FromLong(self->ev.pbutton.button);
    } 
    break;

  case evExpose:
    if (strcmp(name,"x")==0) {
      return PyInt_FromLong(self->ev.expose.x);
    } 
    else if (strcmp(name,"y")==0) {
      return PyInt_FromLong(self->ev.expose.y);
    }
    else if (strcmp(name,"h")==0) {
      return PyInt_FromLong(self->ev.expose.h);
    }
    else if (strcmp(name,"w")==0) {
      return PyInt_FromLong(self->ev.expose.w);
    }    
    break;

  case evCommand:
  case evInformation:
    if (strcmp(name, "code")==0) {
      return PyInt_FromLong(self->ev.cmd.code);
    }
    else if (strcmp(name, "data")==0) {
       // do something snazzy..
       }
    
    break;

  case evValRelative:
  case evValAbsolute:
    // need more info on valuators.
    break;
    */
  default:      
  }

  return 0;
  }
static void event_dealloc(EventObject*self) {
  PyMem_DEL(self);
}


static PyTypeObject event_type = {
  PyObject_HEAD_INIT(NULL)
  0,
  "Event",
  sizeof(EventObject),
  0,
  (destructor)event_dealloc,
  (printfunc)0,
  (getattrfunc)event_getattr,
  (setattrfunc)0,
  (cmpfunc)0,
  (reprfunc)0,
  0,
  0,
  0,
  (hashfunc)0,
  (ternaryfunc)0,
  (reprfunc)0,

  0L,0L,0L,0L,
  "no docs."
};

static PyObject* newEventObjectFromEvent(ggi_event e) {
  EventObject*self;
  self=PyObject_NEW(EventObject, &event_type);
  if(!self) return NULL;
  self->ev=e;
  self->dict=PyDict_New();
  return(PyObject*)self;
}
/* End of EventObject ----------------------------------------------*/








/* VisObject -------------------------------------------------------*/

/*------------ Mode setting --------------------------------*/
static PyObject * vis_setgraphmode(VisObject*self, PyObject*args) 
{
  int i;
  ggi_graphtype gtype;
  int x,y,vx,vy;
 
  if(!(PyArg_ParseTuple(args, "iiiii", &x,&y, &vx, &vy, &gtype))) {
    PyErr_SetString(PyExc_TypeError, "Bad Arguments!");
    return NULL;
  }
  i=ggiSetGraphMode(self->vis, x,y,vx,vy,gtype);
  if(i==0) {
    Py_INCREF(Py_None);
    return Py_None;
  }
  return PyInt_FromLong(i);
}

static PyObject* vis_setmode(VisObject*self, PyObject*args) {
  PyObject*mode;
  int i;
#define mode_ ((ModeObject*)mode)

  if((!PyArg_ParseTuple(args,"O",&mode)) || (!PyMode_Check(mode))) {
    PyErr_SetString(PyExc_TypeError, "Bad Arguments!");
    return NULL;
  }

  i=ggiSetMode(self->vis, &mode_->mode); // !! Modifies mode object !!
  if(i==0) {
    Py_INCREF(Py_None);
    return Py_None;
  }

  return PyInt_FromLong(i);
#undef mode_
}

static PyObject* vis_getmode(VisObject*self, PyObject*args) {
  ggi_mode m;
  ggiGetMode(self->vis, &m); 
  return (PyObject*)newModeFromGGIMode(m);
}

static PyObject* vis_checkmode(VisObject*self, PyObject*args) {
  PyObject*mode;
  int i;
#define mode_ ((ModeObject*)mode)

  if((!PyArg_ParseTuple(args,"O",&mode)) || (!PyMode_Check(mode))) {
    PyErr_SetString(PyExc_TypeError, "Bad Arguments!");
    return NULL;
  }

  i=ggiCheckMode(self->vis, &mode_->mode); // !! Modifies mode object !!
  if(i!=0) {
    Py_INCREF(Py_None);
    return Py_None;
  }

  return PyInt_FromLong(i);
#undef mode_
}

/*-------------- Color Functions ----------------------------*/

// Color(..) is implemented in the ColorObject section

static PyObject* vis_setcolorfulpal(VisObject*self, PyObject*args) {  
  ggiSetColorfulPalette(self->vis);
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject* vis_gamma(VisObject*self, PyObject*args) {
  int i;
  double r,g,b;

  if(PyArg_ParseTuple(args, "ddd", &r, &g, &b)) {
    i=ggiSetGamma(self->vis, r,g,b);
    if(i==0) {
      Py_INCREF(Py_None);
      return Py_None;
    }
    return PyInt_FromLong(i);
  } else if(PyTuple_Size(/*(PyTuple*)*/args)==0) {
    i=ggiGetGamma(self->vis, &r,&g,&b);
    if(i==0) {
      return Py_BuildValue("ddd", r,g,b);
    }
    Py_INCREF(Py_None);
    return Py_None;
  } else {
    PyErr_SetString(PyExc_TypeError, "Bad Arguments!");
    return NULL;
  }
}

static PyObject* vis_flags(VisObject*self, PyObject*args) {
  int f;

  if(PyArg_ParseTuple(args, "i", &f)) {
    if(ggiSetFlags(self->vis, f)==0) return PyInt_FromLong(1); // return true
    Py_INCREF(Py_None);
    return Py_None;
  } else if (PyTuple_Size(args)==0) {
    return PyInt_FromLong(ggiGetFlags(self->vis));
  }
  PyErr_SetString(PyExc_TypeError, "bad arguments!");
  return NULL;
  //note: to add addflags and removeflags methods to this thing, define a new type that
  // does all this stuff with __call__ and __getattr__
}


/*--------- flush & fillscreen ------------------------------*/
static PyObject* vis_fillscreen(VisObject*self,PyObject*args) 
{
  double r,g,b;
  ColorObject*co;
  ggi_pixel p,op;
  ggi_color c;

  if(PyArg_ParseTuple(args,"ddd",&r,&g,&b)) {
    floatcolor(&c,r,g,b);
    p=ggiMapColor(self->vis, &c);
  } else if (PyArg_ParseTuple(args, "O", &co)) {
    if(!PyColor_Check(co)) {
      PyErr_SetString(PyExc_TypeError, "bad arguments!");
      return NULL;
    }
    if(co->visual==NULL) {
      p=ggiMapColor(self->vis, &(co->colors[0]));
    } else if (co->visual==self->vis) {
      p=co->pixels[0];
    } else {
      // different visuals.
      ColorObject*newcolor=newColorFromColorAndVis(self->vis, co);
      p=newcolor->pixels[0];
      Py_DECREF(newcolor);
    }
  } else {
    PyErr_SetString(PyExc_TypeError, "bad args!");
    return NULL;
  }  

  ggiGetGCForeground(self->vis, &op);
  ggiSetGCForeground(self->vis, p);
  ggiFillscreen(self->vis);
  ggiSetGCForeground(self->vis, op);
 
  Py_INCREF(Py_None);
  return Py_None;  
}

static PyObject* vis_flush(VisObject*self,PyObject*args)
{
  ggiFlush(self->vis);

  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject* vis_flushregion(VisObject*self, PyObject*args) {
  int x1,y1,x2,y2;

  if(!PyArg_ParseTuple(args, "iiii", &x1, &y1, &x2, &y2)) {
    PyErr_SetString(PyExc_TypeError, "bad arguments!");
    return NULL;
  }

  ggiFlushRegion(self->vis, x1,y1, (x2-x1), (y2-y1));

  Py_INCREF(Py_None);
  return Py_None;
  
}


/*------------- GC functions --------------------*/
/*static PyObject* vis_setfore(VisObject*self, PyObject*args)
{
  ggi_pixel p;
  ggi_color c;
  PyObject*ro,*go,*bo,*co;
  double r,g,b;

  if(PyArg_ParseTuple(args,"OOO",&ro,&go,&bo)) {
    if(!pyfloat(ro, &r)) return NULL;
    if(!pyfloat(go, &g)) return NULL;
    if(!pyfloat(bo, &b)) return NULL;
    floatcolor(&c, r,g,b);
    p=ggiMapColor(self->vis, &c);
  }
  else if(PyArm_ParseTuple(args,"O",&co)) {
    if(!PyColor_Check(co)) {
#define co_ ((ColorObject*)co)
      if (co_->visual!=self->vis) {
	PyErr_SetString(PyExc_TypeError, "can't use colors across visuals!");
	return NULL;
      }
      if(co_->visual) {
	p=co_->pixels[0];
      }else {
	ggi_color c=co_->colors[0];
	p=ggiMapColor(self->vis, &c);
      }
#undef co_      
    }
  }
  else {
    PyErr_SetString(PyExc_TypeError, "need (a)three numbers 0..1 or (b) a color object");
    return NULL;
  }

  
  ggiSetGCForeground(self->vis, p);

  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject* vis_getfore(VisObject*self, PyObject*args)
{
  ggi_pixel p;
  double r,g,b;
  ColorObject*newcolor;
    
  ggiGetGCForeground(self->vis, &p);    
  newcolor=PyObject_NEW(ColorObject,&color_type);
  if(newcolor==NULL) return NULL;

  newcolor->visual=self->vis;
  newcolor->pixels=(ggi_pixel*)malloc(1); 
  newcolor->pixels[0]=p;
  
  Py_INCREF(newcolor);
  return (PyObject*)newcolor;
}

static PyObject* vis_setback(VisObject*self, PyObject*args)
{
  ggi_pixel p;
  ggi_color c;
  PyObject*ro,*go,*bo,*co;
  ColorObject*newcolor;
  double r,g,b;

  if(PyArg_ParseTuple(args,"OOO",&ro,&go,&bo)) {
    if(!pyfloat(ro, &r)) return NULL;
    if(!pyfloat(go, &g)) return NULL;
    if(!pyfloat(bo, &b)) return NULL;
    floatcolor(&c, r,g,b);
    p=ggiMapColor(self->vis, &c);
  }
  else if(PyArm_ParseTuple(args,"O",&co)) {
    if(!PyColor_Check(co)) {
#define co_ ((ColorObject*)co)
      if (co_->visual!=self->vis) {
	PyErr_SetString(PyExc_TypeError, "can't use colors across visuals!");
	return NULL;
      }
      if(co_->visual) {
	p=co_->pixels[0];
      }else {
	ggi_color c=co_->colors[0];
	p=ggiMapColor(self->vis, &c);
      }
#undef co_      
    }
  }
  else {
    PyErr_SetString(PyExc_TypeError, "need (a)three numbers 0..1 or (b) a color object");
    return NULL;
  }

  
  ggiSetGCBackground(self->vis, p);

  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject* vis_getback(VisObject*self, PyObject*args)
{
  ggi_pixel p;
  double r,g,b;
  ColorObject*newcolor;
    
  ggiGetGCBackground(self->vis, &p);    
  newcolor=PyObject_NEW(ColorObject,&color_type);
  if(newcolor==NULL) return NULL;

  newcolor->visual=self->vis;
  newcolor->pixels=(ggi_pixel*)malloc(1); 
  newcolor->pixels[0]=p;
  
  Py_INCREF(newcolor);
  return (PyObject*)newcolor;
}
*/

static PyObject* vis_clipping(VisObject*self, PyObject*args) {
  int l,r,t,b,i;
  if(!PyArg_ParseTuple(args, "|iiii", &l, &t, &r, &b)) return NULL;
  if(PyTuple_Size(args)==4) {
    i=ggiSetGCClipping(self->vis, l,t,r,b);
    if(i!=0) {
      Py_INCREF(Py_None);
      return Py_None;
    }
    return PyInt_FromLong(1); //as in, return true.
  } else if(PyTuple_Size(args)==0) {
    i=ggiGetGCClipping(self->vis, &l, &t, &r, &b);
    //how can it fail???
    return Py_BuildValue("iiii", l,t,r,b);
  }
  
  PyErr_SetString(PyExc_TypeError, "bad Arguments!");
  return NULL;
}

/*----------- Drawing Primitives -----------------*/
static PyObject* vis_pixel(VisObject*self, PyObject*args)
{
  int x,y;
  double r,g,b;
  PyObject*oa=NULL,*ob=NULL,*oc=NULL;
  ColorObject *co;
  ggi_pixel p;
  ggi_color c;

  if(!PyArg_ParseTuple(args, "ii|OOO", &x,&y, &oa,&ob,&oc)) {
    return NULL;
  }

  if(oc!=NULL) { // ie, iiOOO
    if((pyfloat(oa, &r) + pyfloat(ob, &g) + pyfloat(oc, &b))!=3) {
      return NULL;
    }
    floatcolor(&c, r,g,b);
    p=ggiMapColor(self->vis,&c);
    ggiPutPixel(self->vis,x,y,p);
    co=newColorFromNumbersAndVisual(oa,ob,oc,self->vis);
    //?
    //Py_XDECREF(oa);
    //Py_XDECREF(ob);
    //Py_XDECREF(oc);
  }
  else if(oa!=NULL && PyColor_Check(oa)) { //ie, iiO    
    co=(ColorObject*)oa;
    if(co->visual==self->vis) {
      p=co->pixels[0];
      ggiPutPixel(self->vis, x, y, p);
      co=newColorFromVisualAndSize(self->vis, 1);
      co->pixels[0]=p;
    } else if (co->visual==NULL) {
      p=ggiMapColor(self->vis, &co->colors[0]);
      ggiPutPixel(self->vis, x, y, p);
      co=newColorFromNumbersAndVisual(PyFloat_FromDouble(co->colors[0].r),
				      PyFloat_FromDouble(co->colors[0].g),
				      PyFloat_FromDouble(co->colors[0].b),
				      self->vis);
    }
    else {
      PyErr_SetString(PyExc_TypeError, "Can't transfer colors between visuals!");
      return NULL;
    }
  } else if(PyTuple_Size(args)==2) { // ie, ii
    //x, y -> getpixel (returns a colorObject)
    ggiGetPixel(self->vis, x, y, &p);
    ggiUnmapPixel(self->vis, p, &c);
    colorfloat(&c,&r, &g, &b);
    oa=PyFloat_FromDouble(r);
    ob=PyFloat_FromDouble(g);
    oc=PyFloat_FromDouble(b);
    co=newColorFromNumbersAndVisual(oa,ob,oc,self->vis);
    Py_XDECREF(oa);
    Py_XDECREF(ob);
    Py_XDECREF(oc);  // is this really nescessary?
  }else {
    PyErr_SetString(PyExc_TypeError, "Bad Arguments!");
    return NULL;
  }
 
  return (PyObject*)co;
}

static PyObject* vis_line(VisObject*self,PyObject*args)
{
  int x1,x2,y1,y2;
  double r,g,b;
  ggi_color c;
  ggi_pixel p, oldp;
  ColorObject *co;
  PyObject *oa=NULL,*ob=NULL,*oc=NULL;

  if(!PyArg_ParseTuple(args, "iiiiO|OO", &x1, &y1, &x2, &y2, &oa, &ob, &oc)) {
    return NULL;
  }

  
  if (oc!=NULL) {
    // x1,y1,x2,y2,r,g,b
    if((pyfloat(oa, &r) + pyfloat(ob, &g) + pyfloat(oc, &b))!=3) {
      return NULL;
    }
    floatcolor(&c, r,g,b);
    p=ggiMapColor(self->vis,&c);
    ggiGetGCForeground(self->vis, &oldp); // preserve GC
    ggiSetGCForeground(self->vis, p);
    ggiDrawLine(self->vis,x1,y1,x2,y2);
    ggiSetGCForeground(self->vis, oldp);
  } else if (oa!=NULL && PyColor_Check(oa)) {
    //x1,y1,x2,y2,colorObject
    co=(ColorObject*)oa;
    if(co->visual==self->vis) {
      ggiGetGCForeground(self->vis, &oldp); // preserve GC
      ggiSetGCForeground(self->vis, co->pixels[0]);
      ggiDrawLine(self->vis,x1,y1,x2,y2);
      ggiSetGCForeground(self->vis, oldp);
    } else if (co->visual==NULL) {
      p=ggiMapColor(self->vis, &co->colors[0]);
      ggiGetGCForeground(self->vis, &oldp); // preserve GC
      ggiSetGCForeground(self->vis, p);
      ggiDrawLine(self->vis,x1,y1,x2,y2);
      ggiSetGCForeground(self->vis, oldp);
    }
    else {
      PyErr_SetString(PyExc_TypeError, "Can't transfer colors between visuals!");
      return NULL;
    }
  }
  else {
    PyErr_SetString(PyExc_TypeError, "Bad Arguments!");
    return NULL;
  }

  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject* vis_vline(VisObject*self,PyObject*args)
{
  int x,y1,y2,i;
  double r,g,b;
  ggi_color c;
  ggi_pixel p;
  ColorObject *co;
  uint8 *index_co;
  int pixel_size=(self->pixelformat->size)/8;
  PyObject *oa=NULL, *ob=NULL, *oc=NULL;
#define rightify_coords if(y1>y2) { i=y1; y1=y2; y2=i; }

  if(!PyArg_ParseTuple(args, "iii|OOO", &x, &y1, &y2, &oa, &ob, &oc)) return NULL;

  if(oc!=NULL) {
    //x,y1,y2,r,g,b 
    if((pyfloat(oa, &r) + pyfloat(ob, &g) + pyfloat(oc, &b))!=3)  return NULL;
    rightify_coords;
    floatcolor(&c, r,g,b);
    p=ggiMapColor(self->vis,&c);
    co=newColorFromVisualAndSize(self->vis, y2-y1);
    if(co==NULL) return NULL;
    ggiSetGCForeground(self->vis, p); // save old fg?
    ggiDrawVLine(self->vis,x,y1,(y2-y1));
    ggiGetVLine(self->vis,x,y1,(y2-y1),co->pixels);
  } else if (oa!=NULL && PyColor_Check(oa)) {
    // x,y1,y2,c
    rightify_coords;
    co=(ColorObject*)oa;
    if(co->length < (y2-y1)) {
      PyErr_SetString(PyExc_TypeError, "ColorObject too small to fill that line");
      return NULL;
    }
    if(co->visual==self->vis) {      
      ggiPutVLine(self->vis,x,y1,(y2-y1),co->pixels);
    } else if (co->visual==NULL) {
      co=newColorFromColorAndVis(co,self->vis);
      ggiPutVLine(self->vis,x,y1,(y2-y1),co->pixels);
    }
    else {
      PyErr_SetString(PyExc_TypeError, "Can't transfer colors between visuals!");
      return NULL;
    }
  } else if (PyTuple_Size(args)==3) {
    rightify_coords;
    co=newColorFromVisualAndSize(self->vis, (y2-y1));
    ggiGetVLine(self->vis, x, y1, (y2-y1), co->pixels);
  } else {
    PyErr_SetString(PyExc_TypeError, "Bad Arguments!");
    return NULL;
  }

  return (PyObject*)co;
#undef rightify_coords
}

static PyObject* vis_hline(VisObject*self,PyObject*args)
{
  int x1,x2,y,i;
  double r,g,b;
  ggi_color c;
  ggi_pixel p;
  ColorObject *co;
  uint8*index_co;
  int pixel_size=(self->pixelformat->size)/8;
  PyObject *oa=NULL, *ob=NULL, *oc=NULL;
#define rightify_coords if(x1>x2) { i=x1; x1=x2; x2=i; }

  if(!PyArg_ParseTuple(args, "iii|OOO", &y, &x1, &x2, &oa, &ob, &oc)) return NULL;

  if(oc!=NULL) {
    //y,x1,x2,r,g,b 
    if((pyfloat(oa, &r) + pyfloat(ob, &g) + pyfloat(oc, &b))!=3)  return NULL;
    rightify_coords;
    floatcolor(&c, r,g,b);
    p=ggiMapColor(self->vis,&c);
    co=newColorFromVisualAndSize(self->vis, x2-x1);
    if(co==NULL) return NULL;
    ggiSetGCForeground(self->vis, p); // save old fg?
    ggiDrawHLine(self->vis,x1,y,(x2-x1));
    ggiGetHLine(self->vis,x1,y,(x2-x1),co->pixels);
  } else if (oa!=NULL && PyColor_Check(oa)) {
    // y,x1,x2,c
    rightify_coords;
    co=(ColorObject*)oa;
    if(co->length < (x2-x1)) {
      PyErr_SetString(PyExc_TypeError, "ColorObject too small to fill that line");
      return NULL;
    }
    if(co->visual==self->vis) {      
      ggiPutHLine(self->vis,x1,y,(x2-x1),co->pixels);
    } else if (co->visual==NULL) {
      co=newColorFromColorAndVis(co,self->vis);
      ggiPutHLine(self->vis,x1,y,(x2-x1),co->pixels);
    }
    else {
      PyErr_SetString(PyExc_TypeError, "Can't transfer colors between visuals!");
      return NULL;
    }
  } else if (PyTuple_Size(args)==3) {
    rightify_coords;
    printf("get hline got y=%i x1=%i x2=%i\n", y, x1, x2);
    co=newColorFromVisualAndSize(self->vis, (x2-x1));
    ggiGetHLine(self->vis,x1, y, (x2-x1), co->pixels);
  } else {
    PyErr_SetString(PyExc_TypeError, "Bad Arguments!");
    return NULL;
  }

  return (PyObject*)co;
#undef rightify_coords
}


static PyObject* vis_box(VisObject*self, PyObject*args) {
  int x1,x2,y1,y2,i;
  double r,g,b;
  ColorObject *co;
  uint8 *index_co;
  int pixel_size=(self->pixelformat->size)/8;
  PyObject*oa=NULL,*ob=NULL,*oc=NULL;
  ggi_color c;
  ggi_pixel p;

#define rightify_coords if(x2<x1) {i=x1; x1=x2; x2=i;}; if(y2<y1){i=y1;y1=y2;y2=i;};

  if(!PyArg_ParseTuple(args, "iiii|OOO", &x1, &y1, &x2, &y2, &oa, &ob, &oc)) 
    return NULL;


  if (oc!=NULL) {
    //x1,y1,x2,y2,r,g,b - > drawbox
    rightify_coords;
    if((pyfloat(oa, &r) + pyfloat(ob, &g) + pyfloat(oc, &b))!=3)  return NULL;
    floatcolor(&c, r,g,b);
    p=ggiMapColor(self->vis,&c);
    co=newColorFromVisualAndSize(self->vis, (x2-x1)*(y2-y1));
    ggiSetGCForeground(self->vis, p);
    ggiDrawBox(self->vis,x1,y1,(x2-x1),(y2-y1));
    ggiGetBox(self->vis,x1,y1,(x2-x1),(y2-y1), co->pixels);
  } else if (oa!=NULL && PyColor_Check(oa)) {
    //x1,y1,x2,y2,colorObject -> PutBox
    co=(ColorObject*)oa;
    rightify_coords;
    if(co->length < (x2-x1)*(y2-y1)) {
      PyErr_SetString(PyExc_TypeError, "ColorObject too small to fill that box");
      return NULL;
    }
    if(co->visual==self->vis) {
      ggiPutBox(self->vis,x1,y1,(x2-x1),(y2-y1), co->pixels);
      Py_INCREF(co);
    } else if (co->visual==NULL) {
      co=newColorFromColorAndVis(co,self->vis);
      ggiPutBox(self->vis,x1,y1,(x2-x1),(y2-y1), co->pixels);
    }
    else {
      PyErr_SetString(PyExc_TypeError, "Can't transfer colors between visuals!");
      return NULL;
    }
  } else if(PyTuple_Size(args)==4) {
    //x1,x2,y1,y2 -> getbox
    rightify_coords;
    co=newColorFromVisualAndSize(self->vis, (x2-x1)*(y2-y1));
    ggiGetBox(self->vis, x1, y1, (x2-x1), (y2-y1), co->pixels);
  } else {
    PyErr_SetString(PyExc_TypeError, "Bad Arguments!");
    return NULL;
  }

  //Py_INCREF(co); unneeded, we don't keep a copy?
  return (PyObject*)co;
#undef rightify_coords
}

/* ----------- panning & buffering ---------------*/
#define framefunc(which) \
static PyObject* vis_ ## which ## frame(VisObject*self, PyObject*args) {\
  int i=-1212;\
  if(!PyArg_ParseTuple(args, "|i", &i)) return NULL;\
  if(i!=-1212) return PyInt_FromLong(ggiSet ## which ## Frame(self->vis, i));\
  return PyInt_FromLong(ggiGet ## which ## Frame(self->vis));\
}

framefunc(Display);
framefunc(Write);
framefunc(Read);

#undef framefunc

static PyObject* vis_origin(VisObject*self, PyObject*args) {
  int x=-1212,y=-1212;
  if(!PyArg_ParseTuple(args, "|ii", &x, &y)) return NULL;
  
  if(x!=-1212) {
    if(ggiSetOrigin(self->vis, x,y)==0) return PyInt_FromLong(1);
    Py_INCREF(Py_None);
    return Py_None;
  }
  ggiGetOrigin(self->vis, &x, &y); // can't fail??
  return Py_BuildValue("ii", x, y);
}




/*---------- Puts ----------------------------*/

static PyObject* vis_puts(VisObject*self,PyObject*args) {
  int x,y;
  PyObject*or,*og,*ob;
  ggi_pixel p;
  ggi_color c;
  double r,g,b;
  char *s;

  if(!PyArg_ParseTuple(args, "iisOOO", &x, &y, &s, &or, &og, &ob)) {
    PyErr_SetString(PyExc_TypeError, "need two integer values and a string (with no null chars!), and 3 color args..");
    return NULL;
  }  
  if((pyfloat(or, &r) + pyfloat(og, &g) + pyfloat(ob, &b))!=3)  return NULL;
  floatcolor(&c, r,g,b);
  p=ggiMapColor(self->vis,&c);
  
  ggiSetGCForeground(self->vis, p);
  ggiPuts(self->vis, x,y, s);

  Py_INCREF(Py_None);
  return Py_None;
}

/*------- (minor) Blitting -------------------------------*/
PyObject *vis_copybox(VisObject*self,PyObject*args) {
  int x,y,w,h,nx,ny,res;
  
  if(!PyArg_ParseTuple(args, "iiiiii", &x,&y,&w,&h,&nx,&ny) )
    return NULL;
  
  res=ggiCopyBox(self->vis, x,y,w,h,nx,ny);
  
  return Py_BuildValue("i", &res);
}

PyObject *vis_crossblit(VisObject*self,PyObject*args) {
    int x,y,w,h,nx,ny,res,nw,nh;
    PyObject*nv;
    if(!PyArg_ParseTuple(args, "iiiiOii", &x,&y,&w,&h,&nv,&nw,&nh))
       return NULL;
    if(PyMode_Check(nv))       
      {
	res=ggiCrossBlit(self->vis, x,y,w,h, ((VisObject*)nv)->vis, nw,nh);
	return Py_BuildValue("i", &res);    
      } else {
      PyErr_SetString(PyExc_TypeError, "need a visual object.");
      return NULL;
    }
}

/*-------- Events -----------------------------------------------*/

PyObject* vis_eventread(VisObject*self, PyObject*args) {
  struct timeval *t;
  gii_event_mask mask=ggiGetEventMask(self->vis); // i shouldn't have to do this!
  gii_event *e=(gii_event*)malloc(sizeof(gii_event));
  
  if(!e) {
    PyErr_SetString(PyExc_SystemError, "not enough mem to grab an event??!");
    return NULL;
  } 

  ggiEventRead(self->vis, e, mask);

  return newEventObjectFromEvent(*e);
}

PyObject* vis_eventpoll(VisObject*self,PyObject*args) {
  // on hold pending the help of c.l.p in the timeval arena..
}

PyObject* vis_eventsqueued(VisObject*self,PyObject*args) {
  // no docs on this... assuming it returns the number of events matching the mask are waiting
  // in the queue??
  int mask;

  if(!PyArg_ParseTuple(args, "i", &mask)) return NULL;
  return Py_BuildValue("i", ggiEventsQueued(self->vis, mask));  
}

PyObject* vis_eventmask(VisObject*self,PyObject*args) {
  int mask;
  PyObject*o;


  if(PyTuple_Size(args)==1) {
    o=PyTuple_GetItem(args, 0);
    if(!PyInt_Check(o)) {
      PyErr_SetString(PyExc_TypeError, "Integer required.");
      return NULL;
    }
    mask = PyInt_AsLong(o);
    //set    
    ggiSetEventMask(self->vis, mask);
  }
  else {
    //get
    mask=(int)ggiGetEventMask(self->vis);
  }

  return  Py_BuildValue("i", mask);
}

PyObject* vis_eventsend(VisObject*self,PyObject*args) {
  // on hold pending mutablation of EventObject
}



static struct PyMethodDef vis_methods[] = {
  /*------- Mode Setting/Getting/Checking and utils ---------*/
  {"Mode",         (PyCFunction)ModeConstruct,    METH_VARARGS|METH_KEYWORDS, "no docs"},
  {"setMode",      (PyCFunction)vis_setmode,      1, "no docs"},
  {"getMode",      (PyCFunction)vis_getmode,      0, "no docs"},
  {"checkMode",    (PyCFunction)vis_checkmode,    1, "no docs"},
  {"setGraphMode", (PyCFunction)vis_setgraphmode, 5, "no docs"},
  {"flags",        (PyCFunction)vis_flags,     1, "no docs"},
  
  /*-------- Color utilities ------------------*/
  {"Color",    (PyCFunction)ColorConstruct, METH_VARARGS, "no docs"},
  {"setColorFulPalette", (PyCFunction)vis_setcolorfulpal, 0, "no docs"},
  {"gamma",    (PyCFunction)vis_gamma,      METH_VARARGS, "no docs"},
  
  /*-------- Misc. screen management ------------*/
  {"fillScreen",    (PyCFunction)vis_fillscreen,  3, "no docs"},
  {"flush",         (PyCFunction)vis_flush,       0, "no docs"},
  {"flushRegion",   (PyCFunction)vis_flushregion, 4, "no docs"},

  /*--- GC Functions... almost useless except for clipping  ---*/
  //{"foreground",  (PyCFunction)vis_setfore,   3, "no docs"},
  //{"backgound",   (PyCFunction)vis_setback,   3, "no docs"},
  {"clipping",    (PyCFunction)vis_clipping,   4, "no docs"},

  /*------- panning & buffering ---------------------------*/
  {"displayFrame",  (PyCFunction)vis_Displayframe, METH_VARARGS, "no docs"},
  {"readFrame",     (PyCFunction)vis_Readframe, METH_VARARGS, "no docs"},
  {"writeFrame",    (PyCFunction)vis_Writeframe, METH_VARARGS, "no docs"},  
  {"origin",        (PyCFunction)vis_origin,       METH_VARARGS, "no docs"},


  /*----------- Drawing Primitives -----------------*/
  {"pixel",         (PyCFunction)vis_pixel,     METH_VARARGS, "no docs"},
  {"line",          (PyCFunction)vis_line,      4, "no docs"},
  {"hline",         (PyCFunction)vis_hline,     METH_VARARGS, "no docs"},
  {"vline",         (PyCFunction)vis_vline,     METH_VARARGS, "no docs"},
  {"box",           (PyCFunction)vis_box,       METH_VARARGS, "no docs"},

  /*------------ text output, all alone! ------------------*/
  {"puts",          (PyCFunction)vis_puts,      6, "no docs"},

  /*------------- blitting -----------------------------*/  
  {"copyBox", (PyCFunction)vis_copybox, 6, "no docs"},
  {"crossBlit", (PyCFunction)vis_crossblit, 7, "no docs"},


  /*------------- Events ------------------------*/
  {"eventRead", (PyCFunction)vis_eventread, 1, "no docs"},
  {"eventPoll", (PyCFunction)vis_eventpoll, 2, "no docs"},
  {"eventsQueued", (PyCFunction)vis_eventsqueued, 1, "no docs"},
  {"eventMask", (PyCFunction)vis_eventmask, METH_VARARGS, "no docs"},
  {"eventSend", (PyCFunction)vis_eventsend, 2, "no docs"},
  {NULL, NULL}
};

static void vis_dealloc(VisObject *self) 
{
  ggiClose(self->vis);
  PyMem_DEL(self);
}

static PyObject* vis_getattr(VisObject*self, char*name) 
{
  return Py_FindMethod(vis_methods, (PyObject*)self, name);
}


static PyTypeObject vis_type = {
  PyObject_HEAD_INIT(NULL)
  0,
  "Visual",
  sizeof(VisObject),
  0,
  (destructor)vis_dealloc,
  (printfunc)0,
  (getattrfunc)vis_getattr,
  (setattrfunc)0,
  (cmpfunc)0,
  (reprfunc)0,
  0,
  0,
  0,
  (hashfunc)0,
  (ternaryfunc)0,
  (reprfunc)0,

  0L,0L,0L,0L,
  "no docs."
};


// begin-bad-hack
typedef ggi_visual_t _varidx(char**);
typedef _varidx*varidx;
#define v_ ggi_visual_t
v_ open0(char**v){return ggiOpen(NULL);} 
v_ open1(char**v){return ggiOpen(v[0],NULL);} 
v_ open2(char**v){return ggiOpen(v[0],v[1],NULL);} 
v_ open3(char**v){return ggiOpen(v[0],v[1],v[2],NULL);} 
v_ open4(char**v){return ggiOpen(v[0],v[1],v[2],v[3], NULL);} 
v_ open5(char**v){return ggiOpen(v[0],v[1],v[2],v[3],v[4],NULL);} 
v_ open6(char**v){return ggiOpen(v[0],v[1],v[2],v[3],v[4],v[5],NULL);} 
v_ open7(char**v){return ggiOpen(v[0],v[1],v[2],v[3],v[4],v[5],v[6],NULL);} 
v_ open8(char**v){return ggiOpen(v[0],v[1],v[2],v[3],v[4],v[5],v[6],v[7],NULL);}
v_ open9(char**v){return ggiOpen(v[0],v[1],v[2],v[3],v[4],v[5],v[6],v[7],v[8],NULL);}
v_ open10(char**v){return ggiOpen(v[0],v[1],v[2],v[3],v[4],v[5],v[6],v[7],v[8],v[9],NULL);}

#define kcah(n) &open ## n
varidx openWithVisuals[] ={kcah(0),kcah(1),kcah(2),kcah(3),kcah(4),kcah(5),
			   kcah(6),kcah(7),kcah(8),kcah(9),kcah(10)};
#undef v_
#undef kcah
// bad-hacks-never-end
static PyObject* newVisObject(PyObject*targets) {
  VisObject *self;
  char *visuals[10];
  PyObject *t;
  int i,size;
  
  self=PyObject_NEW(VisObject, &vis_type);
  if(!self) return NULL;

  size=PyTuple_Size(targets);
  if(size>10) {
    PyErr_SetString(PyExc_TypeError,"can only manage 10 arguments.");
    return NULL;
  }

  for(i=0;i<size; i++) {
    t=PyTuple_GetItem(targets, i);
    if(!PyString_Check(t)) {
      PyErr_SetString(PyExc_TypeError, "strings only.");
      return NULL;
    }
    visuals[i]=PyString_AsString(t);
  }


  self->vis=openWithVisuals[size](visuals);
  if(self->vis==NULL) {
    PyErr_SetString(PyExc_SystemError, "ggiOpen returned Null!");
    return NULL;
  }

  self->pixelformat=ggiGetPixelFormat(self->vis);

  return (PyObject*)self;
}

/* end of VisObject --------------------------------------------------*/

/* ggiwrapper itself -------------------------------------------------*/


static PyObject* ggi_init(PyObject*self,PyObject*args) {
  int i;
  i=ggiInit();
  return Py_BuildValue("i", i);
}

static PyObject* ggi_exit(PyObject*self,PyObject*args) {
  int i;
  i=ggiExit();
  return Py_BuildValue("i",i);
}

static PyObject* ggi_open(PyObject*self, PyObject*args) {
  return newVisObject(args);
}

static PyObject* ggi_panic(PyObject*self, PyObject*args) {
  char *buf;
  if(!PyArg_ParseTuple(args, "s", &buf)) buf=strdup("Unexpected error occured! \n"
						    "Call your mother!");
  ggiPanic(buf);
  return NULL; // will never be reached, but it shuts the compiler up.
}

static struct PyMethodDef ggi_methods[] = {
  {"ggiInit",  (PyCFunction)ggi_init, 0, "no doc"},
  {"ggiExit",  (PyCFunction)ggi_exit, 0, "no doc"},
  {"ggiOpen",  (PyCFunction)ggi_open, 5, "no doc"},
  {"ggiPanic", (PyCFunction)ggi_panic, 1, "no doc"},
  {"Color", (PyCFunction)ColorConstruct, METH_VARARGS, "no doc"},
  {"Mode",  (PyCFunction)ModeConstruct, METH_VARARGS | METH_KEYWORDS, "no doc"},
  {NULL, NULL}
};

void initggiwrapper() {
  PyObject *m,*d;

  m=Py_InitModule4("ggiwrapper", ggi_methods, "no docs!",
		   (PyObject*)NULL, PYTHON_API_VERSION);

  // add the constants..
  d= PyModule_GetDict(m);
#define intconst(x) PyDict_SetItemString(d, #x, Py_BuildValue("i", x))
#define renintconst(x,y) PyDict_SetItemString(d, #y, Py_BuildValue("i", x))
  // event types
  intconst(evNothing);
  intconst(evCommand);
  intconst(evInformation);
  intconst(evExpose);
  intconst(evKeyPress);
  intconst(evKeyRelease);
  intconst(evKeyRepeat);
  intconst(evPtrRelative);
  intconst(evPtrAbsolute);
  intconst(evPtrButtonPress);
  intconst(evPtrButtonRelease);
  intconst(evValAbsolute);
  intconst(evValRelative);
  intconst(evLast);

  // event mask bits
  intconst(emNothing);
  intconst(emCommand);
  intconst(emInformation);
  intconst(emExpose);
  intconst(emKeyPress);
  intconst(emKeyRelease);
  intconst(emKeyRepeat);
  intconst(emKey);
  intconst(emKeyboard);
  intconst(emPtrRelative);
  intconst(emPtrAbsolute);
  intconst(emPtrButtonPress);
  intconst(emPtrButtonRelease);
  intconst(emPtrMove);
  intconst(emPtrButton);
  intconst(emPointer);
  intconst(emValRelative);
  intconst(emValAbsolute);
  intconst(emValuator);
  intconst(emZero);
  intconst(emAll);

  // misc
  renintconst(GII_EV_ORIGIN_NONE, evOriginNone);
  renintconst(GII_EV_ORIGIN_SENDEVENT, evOriginSendEvent);
  renintconst(GII_EV_TARGET_ALL, evTargetAll);

  renintconst(GII_CMDFLAG_NODATA, cmdflagNoData);
  renintconst(GII_CMDFLAG_PRIVATE, cmdflagPrivate);
  renintconst(GII_CMDFLAG_EXTERNAL, cmdflagExternal);
  renintconst(GII_CMDCODE_EVENTLOST, cmdflagEventLost);

  renintconst(GII_PBUTTON_LEFT, "pbuttonLeft");
  renintconst(GII_PBUTTON_RIGHT, "pbuttonRight");
  renintconst(GII_PBUTTON_MIDDLE, "pbuttonMiddle");

  // valuators not included because they're not documented very well..

  

#undef intconst


  if(PyErr_Occurred()) Py_FatalError("Can't initialize module ggiwrapper");
}
